import 'package:equatable/equatable.dart';
import 'package:syncrow_app/features/menu/model/region_model.dart';
import 'package:syncrow_app/features/menu/model/time_zone_model.dart';

class ProfileState extends Equatable {
  const ProfileState();

  @override
  List<Object> get props => [];
}

class InitialState extends ProfileState {}

class LoadingInitialState extends ProfileState {}

class UpdateState extends ProfileState {
  final List<TimeZone> timeZoneList;

  UpdateState({required this.timeZoneList});
}

class NameEditingState extends ProfileState {
  final bool editName;

  NameEditingState({required this.editName});
}



class FailedState extends ProfileState {
  final String errorMessage;

  const FailedState({required this.errorMessage});

  @override
  List<Object> get props => [errorMessage];
}


class ImageSelectedState extends ProfileState {}

class ChangeImageState extends ProfileState {}

class SaveState extends ProfileState {}

class LoadingSaveState extends ProfileState {}
class RegionsLoadedState extends ProfileState {
  final List<RegionModel> regions;

  const RegionsLoadedState({required this.regions});
}

class TimeZoneLoadedState extends ProfileState {
  final List<TimeZone> timezone;

  const TimeZoneLoadedState({required this.timezone});
}